package eu.jonahbauer.wizard.client.libgdx.screens;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Screen;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.graphics.g2d.TextureAtlas;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.scenes.scene2d.InputEvent;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.scenes.scene2d.ui.Container;
import com.badlogic.gdx.scenes.scene2d.ui.Label;
import com.badlogic.gdx.scenes.scene2d.ui.Skin;
import com.badlogic.gdx.scenes.scene2d.utils.ClickListener;
import com.badlogic.gdx.utils.viewport.ExtendViewport;
import com.badlogic.gdx.utils.viewport.FitViewport;
import eu.jonahbauer.wizard.client.libgdx.GameAtlas;
import eu.jonahbauer.wizard.client.libgdx.WizardGame;
import eu.jonahbauer.wizard.client.libgdx.actors.game.*;
import eu.jonahbauer.wizard.client.libgdx.util.Pair;
import eu.jonahbauer.wizard.common.messages.observer.*;
import eu.jonahbauer.wizard.common.messages.server.GameMessage;
import eu.jonahbauer.wizard.common.messages.server.ServerMessage;
import eu.jonahbauer.wizard.common.model.Card;

import java.util.*;

public class GameScreen implements Screen {
    private final WizardGame game;

    private ExtendViewport extendViewport;
    private FitViewport viewport;
    private Stage stage;
    private Skin skin;
    private TextureAtlas atlas;
    private TextureRegion background;
    public static final float BACKGROUND_WIDTH_PORTION = 0.7f;

    private final UUID[] players;

    //wahrscheinlich eine Map<UUID, Integer> für Sitzpositionen einfügen
    private final Map<UUID, Integer> seats = new HashMap<>();

    private final UUID self;
    private final UUID session;
    private final Map<UUID, String> names;
    private final Map<Integer, Map<UUID, Integer>> scores = new HashMap<>();

    private int round = - 1;
    private final Map<UUID, List<Card>> hands = new HashMap<>();
    private final Map<UUID, Integer> predictions = new HashMap<>();
    private final Map<UUID, List<List<Card>>> tricks = new HashMap<>();
    private Card.Suit trumpSuit;
    private Card trumpCard;
    private Label[] playerNames;

    private int trick;
    private final List<Pair<UUID, Card>> stack = new ArrayList<>();
    private Pair<UUID, UserInputMessage.Action> activePlayer;
    private ChosenCardActor chosenCardActor;

    ClickListener clickListener = new ClickListener() {
        //TODO - Fehler bei Click auf Objekt während der Bewegung - Behebung durch Entfernung des Click-Listeners
        @Override
        public void clicked(InputEvent event, float x, float y) {
            super.clicked(event, x, y);
            var target = event.getTarget();
            if(target instanceof CardActor cardTarget) {
                chosenCardActor = new ChosenCardActor(cardTarget.getCard(), cardTarget.getTexture());
                //TODO falsche Koordinaten
                chosenCardActor.setPosition(cardTarget.getX(), cardTarget.getParent().getY());
                stage.addActor(chosenCardActor);
                target.remove();
                //Todo Message - Server-Antwort abwarten - moveMethode in chosenCardActor aufrufen
            }
        }

    };

    public GameScreen(WizardGame game) {
        this.game = game;

        int playerCount = 6;
        this.session = UUID.randomUUID();
        this.players = new UUID[playerCount];
        this.names = new HashMap<>();

        HashMap<UUID, Integer> score = new HashMap();
        for (int i = 0; i < playerCount; i++) {
            players[i] = UUID.randomUUID();
            names.put(players[i], "Player " + (i + 1));

            predictions.put(players[i], i + 5);
            score.put(players[i], i);
        }
        scores.put(round, score);

        this.self = players[0];

        ArrayList<List<Card>> trickList = new ArrayList<>();
        ArrayList<Card> trick = new ArrayList();
        trick.add(Card.BLUE_3);
        trick.add(Card.YELLOW_1);
        trick.add(Card.GREEN_1);
        trick.add(Card.RED_5);
        trickList.add(trick);
        trickList.add(trick);
        trickList.add(trick);

        for(int i = 0; i < playerCount; i++) {
            tricks.put(players[i], trickList);
            var cards = new ArrayList<Card>(20);
            for(int j = 0; j < 20; j++) {
                cards.add(Arrays.stream(Card.values()).toList().get(j));
            }
            hands.put(players[i], cards);
        }

        //List<Pair<UUID, Card>> stack
        stack.add(new Pair<>(players[0], Card.BLUE_11));
        stack.add(new Pair<>(players[1], Card.RED_1));
        stack.add(new Pair<>(players[2], Card.YELLOW_11));
        stack.add(new Pair<>(players[3], Card.BLUE_10));
        stack.add(new Pair<>(players[4], Card.BLUE_9));
        stack.add(new Pair<>(players[5], Card.BLUE_8));

        determineSeats(players);
    }

    @Override
    public void show() {
        viewport = new FitViewport(WizardGame.WIDTH, WizardGame.HEIGHT);
        extendViewport = new ExtendViewport(WizardGame.WIDTH, WizardGame.HEIGHT);
        stage = new Stage(viewport);
        skin = new Skin(Gdx.files.internal("uiskin.json"));
        atlas = new TextureAtlas(Gdx.files.internal(GameAtlas.$PATH));
        background = atlas.findRegion(GameAtlas.BACKGROUND);

        this.stage.addListener(clickListener);

        var container = new Container<>(new CardsGroup(hands.get(players[0]), atlas).wrap(false).grow());
        container.setPosition(100, 75);
        container.setSize(1200, 200);
        container.fillX();

        InfoTable infoTable = new InfoTable(skin, players);
        infoTable.fill(names, predictions, tricks, scores);
        infoTable.setPosition(WizardGame.WIDTH * 0.85f, WizardGame.HEIGHT * 0.8f);

        trumpCard = Card.BLUE_1;
        TrumpCardActor trumpCardActor = new TrumpCardActor(trumpCard, atlas);
        trumpCardActor.setPosition(WizardGame.WIDTH * 0.85f, WizardGame.HEIGHT * 0.1f);

        this.stage.addListener(clickListener);

        createLabels();
        cardPlayed(stack);

        Gdx.input.setInputProcessor(stage);

        stage.addActor(container);
        stage.addActor(infoTable);
        stage.addActor(trumpCardActor);

        stage.setDebugAll(true);
    }

    @Override
    public void render(float delta) {
        Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);

        //viewport.apply(true);
        extendViewport.apply(true);
        game.batch.setProjectionMatrix(extendViewport.getCamera().combined);
        game.batch.begin();
        float scale = Math.max(
                extendViewport.getWorldWidth() / WizardGame.WIDTH,
                extendViewport.getWorldHeight() / WizardGame.HEIGHT
        );
        game.batch.draw(background, 0,0, scale * WizardGame.WIDTH * BACKGROUND_WIDTH_PORTION, scale * WizardGame.HEIGHT);
        game.batch.end();

        stage.act(delta);
        stage.draw();
    }

    @Override
    public void resize(int width, int height) {
        extendViewport.update(width, height);
        viewport.update(width, height);
    }

    @Override
    public void pause() {}

    @Override
    public void resume() {}

    @Override
    public void hide() {}

    @Override
    public void dispose() {
        if (stage != null) stage.dispose();
        if (atlas != null) atlas.dispose();
        if (skin != null) skin.dispose();
    }

    private void createLabels() {
        //TODO ggf. Labels für Vorhersage und Stichanzahl einbauen
        playerNames = new Label[players.length];
        int count = 0;
        for(int i = 0; i < players.length; i++) {
            playerNames[i] = new Label(names.get(players[i]), skin);
            if(players[i] != self) {
                switch (count) {
                    case 0 -> {
                        playerNames[i].setPosition(10, WizardGame.HEIGHT * 0.5f - 0.5f * playerNames[i].getHeight());
                        stage.addActor(playerNames[i]);
                    }
                    case 1 -> {
                        playerNames[i].setPosition(WizardGame.WIDTH * BACKGROUND_WIDTH_PORTION - playerNames[i].getWidth() - 10, WizardGame.HEIGHT * 0.5f - 0.5f * playerNames[i].getHeight());
                        stage.addActor(playerNames[i]);
                    }
                    case 2 -> {
                        playerNames[i].setPosition(WizardGame.WIDTH * BACKGROUND_WIDTH_PORTION / (players.length - 2) - playerNames[i].getWidth() * 0.5f, WizardGame.HEIGHT * 0.8f - 0.5f * playerNames[i].getHeight());
                        stage.addActor(playerNames[i]);
                    }
                    case 3 -> {
                        playerNames[i].setPosition(2 * WizardGame.WIDTH * BACKGROUND_WIDTH_PORTION / (players.length - 2) - playerNames[i].getWidth() * 0.5f, WizardGame.HEIGHT * 0.85f - 0.5f * playerNames[i].getHeight());
                        stage.addActor(playerNames[i]);
                    }
                    case 4 -> {
                        playerNames[i].setPosition(3 * WizardGame.WIDTH * BACKGROUND_WIDTH_PORTION / (players.length - 2) - playerNames[i].getWidth() * 0.5f, WizardGame.HEIGHT * 0.8f - 0.5f * playerNames[i].getHeight());
                        stage.addActor(playerNames[i]);
                    }
                    default -> throw new IllegalArgumentException();
                }
                count++;
            }
            else {
                playerNames[i].setPosition(WizardGame.WIDTH * 0.35f - 0.5f * playerNames[i].getWidth(), WizardGame.HEIGHT * 0.05f);
                stage.addActor(playerNames[i]);
            }
        }
    }

    private void cardPlayed(List<Pair<UUID, Card>> stack) {
        while(stack.size() > 0) {
            Pair<UUID, Card> pair = stack.get(stack.size() - 1);
            if(pair.first() == self) {
                //TODO Action
                //chosenCardActor.action()
                stack.remove(pair);
            }
            else {
                for(UUID player : players) {
                    if(player == pair.first()) {
                        drawPlayedCard(seats.get(player), pair.second());
                        stack.remove(pair);
                    }
                }
            }
        }
    }

    private void drawPlayedCard(int pos, Card card) {
        PlayedCardActor playedCardActor = new PlayedCardActor(card, atlas);
        switch (pos) {
            case 1 -> playedCardActor.setPositionOne();
            case 2 -> playedCardActor.setPositionTwo();
            case 3 -> playedCardActor.setPositionThree();
            case 4 -> playedCardActor.setPositionFour();
            case 5 -> playedCardActor.setPositionFive();
        }
        stage.addActor(playedCardActor);
    }

    private void determineSeats(UUID[] players) {
        for(int i = 0; i < players.length; i++) {
            if(self == players[i]) {
                seats.put(self, 0);
            }
            else {
                seats.put(players[i], i);
            }
        }
    }

    public void onMessage(ServerMessage serverMessage) {
        if (serverMessage instanceof GameMessage gameMessage) {
            var observerMessage = gameMessage.getObserverMessage();
            if (observerMessage instanceof CardMessage card) {
                stack.add(Pair.of(card.getPlayer(), card.getCard()));
            } else if (observerMessage instanceof TrickMessage trick) {
                stack.clear();
                tricks.computeIfAbsent(trick.getPlayer(), p -> new ArrayList<>()).add(trick.getCards());
            } else if (observerMessage instanceof HandMessage hand) {
                var cards = hands.computeIfAbsent(hand.getPlayer(), p -> new ArrayList<>());
                cards.clear();
                cards.addAll(hand.getHand());
            } else if (observerMessage instanceof ScoreMessage score) {
                var roundScores = scores.computeIfAbsent(round, r -> new HashMap<>());
                roundScores.putAll(score.getPoints());
            } else if (observerMessage instanceof TrumpMessage trump) {
                trumpCard = trump.getCard();
                trumpSuit = trump.getSuit();
                //Todo update TrumpCardActor if trumpSuit != null and trumpCard != null
            } else if (observerMessage instanceof PredictionMessage prediction) {
                predictions.put(prediction.getPlayer(), prediction.getPrediction());
                //TODO update table and label
            } else if (observerMessage instanceof UserInputMessage userInput) {
                activePlayer = Pair.of(userInput.getPlayer(), userInput.getAction());
                if (self.equals(userInput.getPlayer())) {
                    // TODO do something
                }
            } else if (observerMessage instanceof StateMessage state) {
                switch (state.getState()) {
                    case "starting_round" -> {
                        round++;
                        trick = -1;
                        hands.clear();
                        predictions.clear();
                        tricks.clear();
                        trumpSuit = null;
                        trumpCard = null;
                    }
                    case "starting_trick" -> {
                        trick++;
                        stack.clear();
                        activePlayer = null;
                    }
                    case "finished", "error" -> {
                        // TODO do something
                    }
                }
            }
        }
    }
}
